/*
 * Guy Streeter, 31-mar-92
 *
 * 15-may-92 Fix the handling of Octet Strings, so they come out unmodified.
 *
 *  A lot of this code was derived by example from the CMU SNMP package,
 * whose copyright notice follows:
 */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/
#include "EXTERN.h"
#include "perl.h"

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>
#include "snmp.h"
#include "snmp_impl.h"
#include "asn1.h"
#include "snmp_api.h"
#include "snmp_client.h"

extern int snmp_errno;
extern struct tree *Mib;

int snmp_dump_packet = 0;

static int last_status;
static int last_errstat;
static char err_text[1024];

enum usersubs {
    US_snmp_get,
    US_snmp_next,
    US_snmp_set,
    US_snmp_error,
};

int
userinit()
{
    static void usersub();
    static char *filename = "snmp.c";

    make_usub("snmp_get", US_snmp_get, usersub, filename);
    make_usub("snmp_next", US_snmp_next, usersub, filename);
    make_usub("snmp_set", US_snmp_set, usersub, filename);
    make_usub("snmp_error", US_snmp_error, usersub, filename);
    /* ../perl.h declares this routine as returning an int, but the
       documentation doesn't define the return value, and the code
       doesn't check it */
    return 0;
}


static char *
snmp_error()
{
    if (err_text[0] != '\0') return err_text;

    if (last_status == STAT_SUCCESS) return snmp_errstring(last_errstat);

    if (last_status == STAT_TIMEOUT) return "Timeout ocurred";

    return "A general error ocurred";
}


static int
usersub(ix, sp, items)
  int ix;
  register int sp;
  register int items;
{
    STR **st = stack->ary_array + sp;
    struct snmp_session session, *ss;
    struct snmp_pdu *pdu, *response = NULL;
    oid name[MAX_NAME_LEN];
    int name_length;
    char buf[1024];
    int i;

    switch (ix)
    {
      case US_snmp_get:
      case US_snmp_next:
      case US_snmp_set:
      {
	  char *host = (char *)str_get(st[1]);
	  char *community = (char *)str_get(st[2]);

	  last_status = 0;
	  err_text[0] = '\0';

	  if (Mib == (struct tree *)0) init_mib();

	  bzero((char *)&session, sizeof(struct snmp_session));
	  session.peername = host;
	  session.community = community;
	  session.community_len = strlen(community);
	  session.retries = SNMP_DEFAULT_RETRIES;
	  session.timeout = SNMP_DEFAULT_TIMEOUT;
	  session.authenticator = NULL;

	  snmp_synch_setup(&session);
	  ss = snmp_open(&session);
	  if (ss == NULL)
	  {
	      strcpy(err_text,"Couldn't open snmp");
	      return NULL;
	  }

	  if (ix == US_snmp_set) 
	  {
	      struct variable_list *vars = NULL;

	      if (items < 5)
		fatal(
	   "Usage: &snmp_set($host, $community, $variable, $type, $value...)");
	      pdu = snmp_pdu_create(SET_REQ_MSG);

	      for (i = 3; i <= items; )
	      {
		  struct variable_list *vp;
		  char *type, *variable = (char *)str_get(st[i]);

		  name_length = MAX_NAME_LEN;
		  if ( !read_objid(variable, name, &name_length))
		  {
		      sprintf(err_text,"Invalid object identifier: %s",
			      variable);
		      snmp_free_pdu(pdu);
		      snmp_close(ss);
		      return NULL;
		  }

		  if ((vp = (struct variable_list *)malloc(sizeof *vp))
		      == NULL) fatal ("malloc() failed, line %d",__LINE__);
		  if (vars == NULL)
		    pdu->variables = vp;
		  else
		    vars->next_variable = vp;
		  vars = vp;

		  vp->name_length = name_length;
		  if ((vp->name = (oid *)malloc(vp->name_length * sizeof(oid)))
		      == NULL) fatal("malloc() failed, line %d",__LINE__);
		  bcopy((char *)name, (char *)vp->name,
			vp->name_length * sizeof(oid));

		  type = (char *)str_get(st[++i]);
		  variable = (char *)str_get(st[++i]);

		  if (strcmp(type, "INTEGER") == 0)
		  {
		      vp->type = INTEGER;
		      if ((vp->val.integer = (long *)malloc(sizeof(long)))
			  == NULL) fatal("malloc() failed");
		      vp->val_len = sizeof(long);
		      *(vp->val.integer) = atoi(variable);
		  }
		  else if (strcmp(type, "DisplayString") == 0)
		  {
		      vp->type = STRING;
		      if ((vp->val.string = strdup(variable)) == NULL)
			fatal("strdup() failed");
		      vp->val_len = strlen(variable);
		  }
		  else if (strcmp(type, "OCTET STRING") == 0)
		  {
		      vp->type = STRING;
		      vp->val_len = st[i]->str_cur;
		      if ((vp->val.string = (char *)malloc(vp->val_len))
			  == NULL) fatal("malloc() failed");
		      bcopy(st[i]->str_ptr, vp->val.string, vp->val_len);
		  }
		  else if (strcmp(type, "NULL") == 0)
		  {
		      vp->type = NULLOBJ;
		      vp->val_len = 0;
		      vp->val.string = NULL;
		  }
		  else if (strcmp(type, "OBJECT IDENTIFIER") == 0)
		  {
		      vp->type = OBJID;
		      vp->val_len = MAX_NAME_LEN;
		      read_objid(variable, name, &vp->val_len);
		      vp->val_len *= sizeof(oid);
		      if ((vp->val.objid = (oid *)malloc(vp->val_len))
			  == NULL) fatal("malloc() failed");
		      bcopy((char *)name, (char *)vp->val.objid,
			    vp->val_len);
		  }
		  else if (strcmp(type, "TimeTicks") == 0)
		  {
		      vp->type = TIMETICKS;
		      if ((vp->val.integer = (long *)malloc(sizeof(long)))
			  == NULL) fatal("malloc() failed");
		      vp->val_len = sizeof(long);
		      *(vp->val.integer) = atoi(variable);
		  }
		  else if (strcmp(type, "IpAddress") == 0)
		  {
		      vp->type = IPADDRESS;
		      if ((vp->val.integer = (long *)malloc(sizeof(long)))
			  == NULL) fatal("malloc() failed");
		      vp->val_len = sizeof(long);
		      *(vp->val.integer) = inet_addr(variable);
		  }
		  else if (strcmp(type, "Counter") == 0)
		  {
		      vp->type = COUNTER;
		      if ((vp->val.integer = (long *)malloc(sizeof(long)))
			  == NULL) fatal("malloc() failed");
		      vp->val_len = sizeof(long);
		      *(vp->val.integer) = atoi(variable);
		  }
		  else if (strcmp(type, "Gauge") == 0)
		  {
		      vp->type = GAUGE;
		      if ((vp->val.integer = (long *)malloc(sizeof(long)))
			  == NULL) fatal("malloc() failed");
		      vp->val_len = sizeof(long);
		      *(vp->val.integer) = atoi(variable);
		  }
		  else if (strcmp(type, "Opaque") == 0)
		  {
		      vp->type = OPAQUE;
		      vp->val_len = st[i]->str_cur;
		      if ((vp->val.string = (char *)malloc(vp->val_len))
			  == NULL) fatal("malloc() failed");
		      bcopy(st[i]->str_ptr, vp->val.string, vp->val_len);
		  }
		  else
		    fatal(
		     "snmp_set() bad type: use one of INTEGER, Counter, etc.");

		  i++;
	      }
	  }
	  else 
	  {
	      if (items < 3)
		fatal("Usage: &snmp_%s($host, $community, $variable, ...)",
		      ix == US_snmp_get ? "get" : "next");
	      pdu = snmp_pdu_create(ix == US_snmp_get
				    ? GET_REQ_MSG : GETNEXT_REQ_MSG);

	      for (i = 3; i <= items; i++)
	      {
		  char *variable = (char *)str_get(st[i]);

		  name_length = MAX_NAME_LEN;
		  if ( !read_objid(variable, name, &name_length))
		  {
		      sprintf(err_text,"Invalid object identifier: %s",
			      variable);
		      snmp_free_pdu(pdu);
		      snmp_close(ss);
		      return NULL;
		  }

		  snmp_add_null_var(pdu, name, name_length);
	      }
	  }

retry:	    
	  last_status = snmp_synch_response(ss, pdu, &response);

	  i = 0;
	  if (last_status == STAT_SUCCESS)
	  {
	      last_errstat = response->errstat;
	      if (response->errstat == SNMP_ERR_NOERROR)
	      {
		  struct variable_list *vars;

		  for (vars = response->variables;
		       vars != NULL;
		       vars = vars->next_variable)
		  {
		      i++;
		      astore(stack, sp + i*2, Nullstr);
		      st = stack->ary_array + sp + (i-1)*2;
		      buf[0] = '.';
		      get_symbol(vars->name,vars->name_length, Mib, buf+1);
		      st[0] = str_2mortal(str_make(buf,strlen(buf)));

		      if (vars->type == ASN_OCTET_STR)
		      {
			  if (vars->val_len == 0)
			    st[1] = str_2mortal(&str_undef);
			  else
			    st[1] = str_2mortal(str_make(vars->val.string,
							 vars->val_len));
		      }
		      else 
		      {
			  sprint_value(buf, vars->name, vars->name_length,
				       vars);
			  if (buf[0] == '\0')
			    st[1] = str_2mortal(&str_undef);
			  else
			    st[1] = str_2mortal(str_make(buf,strlen(buf)));
		      }
		  }

	      }
	      else
		if (ix != US_snmp_set &&
		    (pdu = snmp_fix_pdu(response, ix == US_snmp_get
					? GET_REQ_MSG : GETNEXT_REQ_MSG))
		    != NULL)
		  goto retry;
	  }

	  if (response) snmp_free_pdu(response);

	  snmp_close(ss);

	  err_text[0] = '\0';

	  return sp + i*2 - 1;
      }

      case US_snmp_error:
	if (items != 0)
	  fatal("Usage: &snmp_error()");
	else
	  str_set(st[0], (char*) snmp_error());
	return sp;
    }
    fatal("Can't happen");
}
